"""
Unit tests for gists_gone.py
"""

from datetime import datetime
import json
from pathlib import Path
from unittest.mock import patch, Mock

import pytest

from requests import HTTPError

from gists_gone.gists_gone import (
    cli,
    parse_date_arguments,
    create_gists,
    filter_gists,
    delete_gists,
    Gist,
)


# Fixtures.
@pytest.fixture
def mock_gists():
    with open(
        Path("./tests/fixtures/fake_gists.json"), "r", encoding="utf-8"
    ) as json_input:
        gists = json.load(json_input)
    return gists


@pytest.fixture
def created_gists(mock_gists):
    gists = create_gists(mock_gists)
    return gists


# Test CLI.
@patch("gists_gone.gists_gone.get_parser_args")
@patch.dict("gists_gone.gists_gone.environ", {}, clear=True)
def test_no_api_key(cli_args):
    """Test a ValueError is raised when no API token is passed to the
    --token option AND no value is found with GITHUB_API_TOKEN environmental
    variable.
    """
    cli_args.return_value = Mock(token=None)
    with pytest.raises(ValueError):
        cli()


@patch("gists_gone.gists_gone.get_parser_args")
@patch.dict("gists_gone.gists_gone.environ", {"GITHUB_API_TOKEN": "123"})
def test_invalid_api_key(cli_args, requests_mock):
    """Test an exception is raised when an invalid API token is provided
    to the CLI.
    """
    cli_args.return_value = Mock(token=None)
    requests_mock.get("https://api.github.com/gists", status_code=403)

    with pytest.raises(HTTPError):
        cli()


@patch("gists_gone.gists_gone.create_gists")
@patch("gists_gone.gists_gone.filter_gists")
@patch("gists_gone.gists_gone.delete_gists")
@patch.dict("gists_gone.gists_gone.environ", {"GITHUB_API_TOKEN": "123"})
def test_no_filtering(mock_delete, mock_filter, gists, requests_mock):
    """Test that no filtering is applied when no arguments have been passed
    to the CLI and that delete_gists is called.

    Notes
    -----
    Unit testing this method requires swapping out the Mock object
    for a custom class, as calling vars() on a Mock returns attributes
    not present in the object I'm trying to mock.

    Note that this is also why I have used the patch decorator in the body of
    this unit test, as I needed to define swap out the mock for a custom class
    defined within this function.
    """

    class Arguments:
        def __init__(
            self,
            token=None,
            force=False,
            visibility=None,
            languages=None,
            date_range=None,
        ):
            self.token = token
            self.force = force
            self.visibility = visibility
            self.languages = languages
            self.date_range = date_range

    requests_mock.get("https://api.github.com/gists", status_code=200, json={})
    with patch("gists_gone.gists_gone.get_parser_args", Arguments):
        cli()
        mock_filter.assert_not_called()
        mock_delete.assert_called()


@patch("gists_gone.gists_gone.create_gists")
@patch("gists_gone.gists_gone.filter_gists")
@patch("gists_gone.gists_gone.delete_gists")
@patch.dict("gists_gone.gists_gone.environ", {"GITHUB_API_TOKEN": "123"})
def test_filtering_is_called(mock_delete, mock_filter, gists, requests_mock):
    """Test that the filter_gists is called onlt when a user passes an argument
    to the CLI.

    Notes
    -----
    See test_no_filtering() documentation for details on set up.
    """

    class Arguments:
        def __init__(
            self,
            token=None,
            force=None,
            visibility=False,
            languages=None,
            date_range=None,
        ):
            self.token = token
            self.force = force
            self.visibility = visibility
            if languages is None:
                self.languages = ["Python", "SQL"]
            else:
                self.languages = languages
            self.date_range = date_range

    requests_mock.get("https://api.github.com/gists", status_code=200, json={})
    with patch("gists_gone.gists_gone.get_parser_args", Arguments):
        cli()
        mock_filter.assert_called()
        mock_delete.assert_called()


# Test functions.
def test_parse_date_arguments():
    """Test that the test_parse_date_arguments works."""
    # Test that too many arguments raises an error.
    arguments = ["2024-01-01", "2024-02-01", "2024-03-01"]
    with pytest.raises(TypeError):
        parse_date_arguments(arguments)
    # Test that a list is returned.
    arguments = ["2024-01-01"]
    assert isinstance(parse_date_arguments(arguments), list)
    # Test that datetime objects are succesfully created.
    assert parse_date_arguments(arguments)[0] == datetime(2024, 1, 1).date()
    arguments = ["2024-01-01", "2024-02-01"]
    assert parse_date_arguments(arguments)[0] == datetime(2024, 1, 1).date()
    assert parse_date_arguments(arguments)[1] == datetime(2024, 2, 1).date()
    # Test that an exception is raised when a string in the wrong format is passed.
    arguments = ["2024-01"]
    with pytest.raises(ValueError):
        parse_date_arguments(arguments)
    arguments = ["2024-01-01", "2024"]
    with pytest.raises(ValueError):
        parse_date_arguments(arguments)


@patch("gists_gone.gists_gone.get_parser_args")
@patch.dict("gists_gone.gists_gone.environ", {"GITHUB_API_TOKEN": "123"})
def test_create_gists_works(cli_args, requests_mock, mock_gists):
    """Test that Gists named tuples are succesfully created from the JSON
    returned from the Github API."""
    requests_mock.get("https://api.github.com/gists", status_code=200)
    gists = create_gists(mock_gists)

    assert isinstance(gists, list)
    assert gists[0] == Gist(
        "7fea2e3837f324e5e3699917f687c862",
        "secret",
        "Clojure",
        datetime(2024, 7, 12).date(),
    )
    assert gists[2] == Gist(
        "3a9f7f73665cf174f9466e3f28fcaf89",
        "public",
        "Unknown",
        datetime(2024, 7, 10).date(),
    )


def test_filter_gists_returns_list(created_gists):
    """Test that filter_gists() returns a list."""
    arguments = [None, ["Python"], None]
    gist_ids = filter_gists(arguments, created_gists)
    assert isinstance(gist_ids, list)


def test_filter_gists_works_with_visibility(created_gists):
    """Test that filter_gists() works when the visibility argument is passed."""

    arguments = ["secret", None, None]
    gist_ids = filter_gists(arguments, created_gists)
    assert "7fea2e3837f324e5e3699917f687c862" in gist_ids
    assert "5f6258f9caae6f2c6511e926f7f623af" in gist_ids
    assert len(gist_ids) == 2

    arguments = ["public", None, None]
    gist_ids = filter_gists(arguments, created_gists)
    assert "7fea2e3837f324e5e3699917f687c862" not in gist_ids
    assert "5f6258f9caae6f2c6511e926f7f623af" not in gist_ids
    assert len(gist_ids) == 4


def test_filter_gists_works_with_languages(created_gists):
    """Test that filter_gists() works when a languages argument is passed."""

    # One language passed to --languages.
    arguments = [None, ["Python"], None]
    gist_ids = filter_gists(arguments, created_gists)
    assert "68ae668e6b5e7364f44b62dd7062231f" in gist_ids
    assert len(gist_ids) == 1

    # Multiple language passed to --languages.
    arguments = [None, ["Python", "SQL"], None]
    gist_ids = filter_gists(arguments, created_gists)
    assert "68ae668e6b5e7364f44b62dd7062231f" in gist_ids
    assert "bc22d164463296d99cbeb1a7038b6d6e" in gist_ids
    assert len(gist_ids) == 2

    # Multiple arguments passed to --languages, only one of which is valid.
    arguments = [None, ["Python", "Spam"], None]
    gist_ids = filter_gists(arguments, created_gists)
    assert "68ae668e6b5e7364f44b62dd7062231f" in gist_ids
    assert len(gist_ids) == 1


def test_filter_works_with_dates(created_gists):
    """Test that filter_gists() works with arguments passed to -dr."""

    # Test Gist creation date equal to date passed to argument.
    arguments = [None, None, ["2024-06-16"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 3
    assert "8eaee095f4b3a822127cc4fa368b4165" in gist_ids
    assert "bc22d164463296d99cbeb1a7038b6d6e" in gist_ids
    assert "68ae668e6b5e7364f44b62dd7062231f" in gist_ids

    # Test Gist created within date ranges.
    arguments = [None, None, ["2024-07-11", "2024-08-01"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 2

    arguments = [None, None, ["2021-01-01", "2023-01-01"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 0

    arguments = [None, None, ["2025-01-01", "2030-01-01"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 0


def test_filter_works_with_multiple_arguments(created_gists):
    """Test that filter_gists works when different arguments are passed."""

    arguments = ["secret", ["Python"], None]
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 0

    arguments = ["public", ["Python"], None]
    gist_ids = filter_gists(arguments, created_gists)
    assert "68ae668e6b5e7364f44b62dd7062231f" in gist_ids
    assert len(gist_ids) == 1

    arguments = ["public", ["Python", "Ruby"], None]
    gist_ids = filter_gists(arguments, created_gists)
    assert "68ae668e6b5e7364f44b62dd7062231f" in gist_ids
    assert "8eaee095f4b3a822127cc4fa368b4165" in gist_ids
    assert len(gist_ids) == 2

    arguments = ["secret", None, ["2024-07-12"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 2

    arguments = ["public", ["Rust", "Clojure"], ["2024-07-12"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 0

    arguments = ["secret", ["Rust", "Clojure"], ["2024-07-12"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 2

    arguments = ["public", None, ["2024-04-01", "2024-06-28"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 3

    arguments = ["public", ["Ruby"], ["2024-04-01", "2024-06-28"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 1
    assert "8eaee095f4b3a822127cc4fa368b4165" in gist_ids

    arguments = ["public", ["SQL"], ["2024-01-01", "2024-06-15"]]
    arguments[2] = parse_date_arguments(arguments[2])
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 0

    # Technically filter_gists() should not be called if all arguments are None.
    arguments = [None, None, None]
    gist_ids = filter_gists(arguments, created_gists)
    assert len(gist_ids) == 6


@patch("builtins.input")
@patch("requests.delete")
def test_force_argument_works(mock_api_call, mock_input, created_gists):
    """
    Test that the --force option works with delete_gists.
    """
    delete_gists(Mock(force=True), created_gists)
    mock_input.assert_not_called()


@patch("builtins.input")
@patch("gists_gone.gists_gone.sleep")
@patch("requests.delete")
def test_delete_returns_if_users_says_no(mock_api_call, sleep, answer, created_gists):
    """Test that if a user does not respond with Yes, Y or yes the delete_gists()
    function will exit without doing anything.
    """
    answer.return_value = "No"
    delete_gists(Mock(force=False), created_gists)
    sleep.assert_not_called()


@patch("builtins.input")
@patch("gists_gone.gists_gone.sleep")
@patch("requests.delete")
def test_delete_does_not_run_with_no_gists(mock_api_call, sleep, mock_input):
    """
    Test that if no gists have been passed to delete_gists() nothing happens.
    """
    mock_input.assert_not_called()
    sleep.assert_not_called()
    delete_gists(Mock(force=False), {})



"""
gists-gone is a CLI that gives you more granular control over bulk deletion of
your Github gists.
"""

import argparse
from collections import namedtuple
from datetime import datetime
from os import environ
from time import sleep

from alive_progress import alive_bar
import requests
import rich

Gist = namedtuple(
    "Gist",
    ["id", "visibility", "language", "created_date"],
)


def get_parser_args():
    """Create the command-line parser and get arguments."""
    parser = argparse.ArgumentParser(
        description="Bulk delete Github Gists from the command-line."
    )
    parser.add_argument(
        "-t", "--token", type=str, help="Your Github API access token.", required=False
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Don't ask for confirmation before deletion of gists. Use with caution.",
        required=False,
    )

    parser.add_argument(
        "-v",
        "--visibility",
        type=str,
        help="Specify whether you want to delete public or secret gists.",
        required=False,
        choices=["public", "secret"],
    )

    parser.add_argument(
        "-l",
        "--languages",
        type=str,
        help="Specify gists in which which programming languages to delete. Default is all.",
        nargs="+",
        required=False,
    )

    parser.add_argument(
        "-dr",
        "--date_range",
        type=str,
        help="Specify the date or date range gists were created.",
        nargs="+",
        required=False,
    )

    args = parser.parse_args()
    return args


def cli():
    # Check for access token.
    args = get_parser_args()
    if not args.token:
        try:
            args.token = environ["GITHUB_API_TOKEN"]
        except KeyError:
            raise ValueError(
                "Please pass your Github API token to the --token option or create an environmental variable."
            )

    relevant_args = [
        value for key, value in vars(args).items() if key not in ("token", "force")
    ]
    # Process date arguments.
    if relevant_args[2] is not None:
        relevant_args[2] = parse_date_arguments(relevant_args[2])

    # Verify token is valid and get all gists.
    gists = []
    for page in range(1, 31):
        response = requests.get(
            "https://api.github.com/gists",
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {args.token}",
                "X-GitHub-Api-Version": "2022-11-28",
            },
            params={"per_page": "100", "page": f"{page}"},
        )
        response.raise_for_status()
        page_gists = create_gists(response.json())
        gists.extend(page_gists)
        if len(page_gists) < 30:  # No gists on further pages.
            break
        sleep(1)  # To avoid secondary rate limits.

    # Filter and delete gists.
    if all(value is None for value in relevant_args):
        gist_ids = [gist.id for gist in gists]
        delete_gists(args, gist_ids)
    else:
        gist_ids = filter_gists(relevant_args, gists)
        delete_gists(args, gist_ids)


def parse_date_arguments(date_arguments):
    """Create a list of datetime objects from the arguments passed to the
    CLI's -dr option."""
    if len(date_arguments) == 3:
        raise TypeError(
            "Too many arguments passed to the -dr flag. Only pass max 2 dates."
        )
    dates = []
    for argument in date_arguments:
        try:
            date = datetime.strptime(argument, "%Y-%m-%d").date()
            dates.append(date)
        except ValueError as error:
            raise ValueError(
                "Please pass a date in YYYY-MM-DD format to the -dr argument."
            ) from error
    return dates


def create_gists(json):
    """Create Gists."""
    gists = []
    for raw_gist in json:
        if (
            languages := raw_gist["files"][next(iter(raw_gist["files"]))]["language"]
        ) is None:
            languages = "Unknown"
        visibility = raw_gist.get("public")
        date_created = datetime.strptime(
            raw_gist["created_at"], "%Y-%m-%dT%H:%M:%SZ"  # ISO 8601 format.
        ).date()
        if visibility:
            visibility = "public"
        else:
            visibility = "secret"

        gist = Gist(raw_gist.get("id"), visibility, languages, date_created)
        gists.append(gist)
    return gists


def filter_gists(relevant_args, gists):
    """Filter the gists to those that match the CLI arguments."""
    gist_ids = []
    for index, argument in enumerate(relevant_args):
        matched_ids = []
        for gist in gists:
            if (
                argument is None
            ):  # All Gists eligible if there is no argument for this option.
                matched_ids.append(gist.id)
                continue
            # Visiblity.
            if index == 0 and argument == gist.visibility:
                matched_ids.append(gist.id)
            # Languages.
            if index == 1 and isinstance(argument, list):
                if gist.language in argument:
                    matched_ids.append(gist.id)
            # Date range
            if index == 2 and isinstance(argument, list):
                if len(argument) == 1:
                    if gist.created_date == argument[0]:
                        matched_ids.append(gist.id)
                else:
                    if (
                        gist.created_date >= argument[0]
                        and gist.created_date <= argument[1]
                    ):
                        matched_ids.append(gist.id)
        gist_ids.append(matched_ids)

    gist_ids = set(gist_ids[0]).intersection(*gist_ids[1:])
    return list(gist_ids)


def delete_gists(args, gists):
    """Delete Github gists."""
    if len(gists) == 0:
        print("No gists are eligible for deletion.")
        return
    if not args.force:
        rich.print(
            """[bold red]Are you [underline]sure[/underline] you proceed with the deletion?[/bold red]"""
        )
        rich.print(
            f"""[bold yellow]{len(gists)}[/bold yellow] [bold red]gists will be deleted.[/bold red]"""
        )
        answer = input("[Y/n] ")
        if answer not in ("Yes", "Y", "y"):
            return
    with alive_bar(len(gists)) as progress_bar:
        progress_bar.title("Deleting gists...")
        for gist in gists:
            sleep(1)
            response = requests.delete(
                f"https://api.github.com/gists/{gist}",
                headers={
                    "Accept": "application/vnd.github+json",
                    "Authorization": f"Bearer {args.token}",
                    "X-GitHub-Api-Version": "2022-11-28",
                },
            )
        response.raise_for_status()
        progress_bar()
    rich.print("[bold red]Gists have been deleted![/bold red]")
